/**
 * Log.c - Error reporting facilities
 * Wilfredo Sanchez | wsanchez@opensource.apple.com
 * $Apple$
 **
 * Copyright (c) 1999-2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Portions Copyright (c) 1999 Apple Computer, Inc.  All Rights
 * Reserved.  This file contains Original Code and/or Modifications of
 * Original Code as defined in and that are subject to the Apple Public
 * Source License Version 1.1 (the "License").  You may not use this file
 * except in compliance with the License.  Please obtain a copy of the
 * License at http://www.apple.com/publicsource and read it before using
 * this file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON- INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 **/

#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <syslog.h>
#include <err.h>
#include <CoreFoundation/CoreFoundation.h>

#include "Log.h"

int initLog()
{
    /* Open log facility */
    openlog(kLogName, LOG_CONS|LOG_ODELAY, LOG_AUTH);

    return(0);
}

#define INIT														\
    CFStringRef aFormattedString;											\
    CFDataRef   aStringData;												\
    va_list anArgs;													\
															\
    va_start(anArgs, aMessage);												\
    aFormattedString = CFStringCreateWithFormatAndArguments(NULL, NULL, aMessage, anArgs);				\
    va_end(anArgs);													\
															\
    aStringData = CFStringCreateExternalRepresentation(NULL, aFormattedString, kCFStringEncodingASCII, '?')

#define FORMAT "%.*s", (int)CFDataGetLength(aStringData), CFDataGetBytePtr(aStringData)

#define RELEASE				\
    CFRelease(aStringData);		\
    CFRelease(aFormattedString)

void debug (CFStringRef aMessage, ...)
{
    INIT;
    syslog(LOG_DEBUG, FORMAT);
    printf(FORMAT);
    RELEASE;
}

void message (CFStringRef aMessage, ...)
{
    INIT;
    syslog(LOG_INFO, FORMAT);
    printf(FORMAT);
    RELEASE;
}

void warning (CFStringRef aMessage, ...)
{
    INIT;
    syslog(LOG_ALERT, FORMAT);
    printf(FORMAT);
    RELEASE;
}

void error (CFStringRef aMessage, ...)
{
    INIT;
    syslog(LOG_ERR, FORMAT);
    printf(FORMAT);
    RELEASE;
}

void emergency (CFStringRef aMessage, ...)
{
    INIT;
    syslog(LOG_CRIT, FORMAT);
    printf(FORMAT);
    RELEASE;
}
